      program nxs2as
c***** reads ILL .nxs file and converts to standard ascii format
c
c      uses dictionary to find parameters
c      D33index.nxf for D33
c      D22index.nxf for D22
c      D11index.nxf for D11
c      these contain dataset absolute names, par-location, multiplier, def. value
c      if found in the current workdir it is used, otherwise file from
c      SAS_DIR is used
c     the dictionary depends on SAS_INSTRUMENT
c     raw data sought fro SAS_INSTRUMENT
c     ascii data written into SAS_CACHE directory for reuse by olde programs
c     this allows the program to be run as a "decompress" utility
c     setting the path to nxs2as  in SAS_DECOMPRESS
c     in the command line version there is greater freedom to define
c     input and output directories.
c      Ron Ghosh, Epsom September 2012
c
c     updated May 2015
      parameter (NXYT=100*128*256)
      integer idon(100*256*128)
      real par(128)
      integer i0(256)
      common/h5work/iv(NXYT)
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasmed/medpar(156)
      common/sasnxc/version
      common/sasinx/instx
      common/nxpars/parnxs(100)
      common/nx2asc/idxfil
      common/pathc/ipath,opath
      common/pathl/lipath,lopath
c***** tests sasas for nxs files
      common/scannxc/igf
      common/scannam/dnames(500)
      common/scannml/nnam,nl(500)
      common/sdbg/ideb
      character nomexp*10,ddate*20
      character dnames*100
      character*1 text(512)
      character*1 term,bsl,ssl
      character lin*80
      character fname*6,instx*4,version*80,cmd*100,idxfil*100
      character cashnam*100,ipath*100,fullnam*100,argf*20,outdir*100
      character opath*100,outfil*110,nxscmd*20,dbg*10
      integer ifound,itran,iraw
      logical found
      data i0/256*0/
      dbg=' '
      ssl='/'
      bsl=char(92)
c****** forward/backslash
      call getenv('SAS_DEBUG',dbg)
      ideb=index(dbg,' ')
      nxscmd=' '
      call getenv('SAS_NXSCMD',nxscmd)
      nxsc=index(' ',nxscmd)
      itran=0
      ifound=0
      iraw=0
      instx=' '
      fullnam=' '
      cashnam=' '
      ipath=' '
      opath=' '
c***** show default is cwd
      lipath=-1
      lopath=-1
      idxfil=' '
      outfil=' '
      outdir=' '
      cmd=' '
c***** use run numbers from...to   in command line and 
c      read from directory SAS_DATA_PATH
c      write to  directory SAS_CACHE
c
c     otherwise prompt for input and output directories
c*****
c      check output directory is writeable.. 
c
      call getenv('SAS_INSTRUMENT',instx)
      if(instx.eq.' ') call stopms('SAS_INSTRUMENT not set')
      if(instx(1:1).eq.'d') instx(1:1)='D'
      idxfil=instx(1:3)//'index.ndf'
      li=12
      narg=iargc()
      if(nxsc.eq.1) narg=0
c***** ignore command line arguments unless SAS_NXSCMD is set (true)
      if(narg.lt.1) go to 5
c***** dont use cache concept in manual input mode
      call getenv('SAS_CACHE',cashnam)
      if(cashnam.eq.' ') 
     1write(*,*) 'nxs2as cmdline needs SAS_CACHE -> manual input'
      if(cashnam.eq.' ') narg=1
      if(cashnam.eq.' ') go to 5
c***** look for cache directory
      lc=len_trim(cashnam)
      inquire(file=cashnam(1:lc),exist=found)
      if(found) go to 66
      cmd='mkdir '//cashnam(1:lc)
      call system(cmd)
      inquire(file=cashnam(1:lc),exist=found)
      if(.not.found) 
     1write(*,*)' Cannot create cache directory ',cashnam(1:lc)
      if(.not.found) call stopms('END')
66    opath=cashnam
      lo=lc
8     irst=0
      last=0
      call getarg(1,argf)
      argf(15:20)=' 0 0/'
      read(argf,*) irst
      call getarg(2,argf)
      argf(15:20)=' 0 0 /'
      read(argf,*) last
      if(opath(lo:lo).ne.'/') then
        lo=lo+1
        opath(lo:lo)='/'
	
      end if	
      go to 29
c***** and treat range first..last

c***** prompt for input
5     write(6,6)
6     format(/' nxs2as - transcribes ILL-SAS NeXus data to ASCII file'/
     1' Version 1.2, Ron Ghosh, Epsom, May 2015'//
     1' Give directory containing nnnnnn.nxs data files : ',$)
      read(5,2) ipath
2     format(a)
      lp=len_trim(ipath)
c***** if lp is 0 then in cwd!
      if(lp.ne.0) then
        lpp=lp
	term=ipath(lpp:lpp)	
	if(term.eq.'/'.or.term.eq.bsl) lpp=lp-1
c***** strip off terminator
        inquire(file=ipath(1:lpp),exist=found)
        if(.not.found) call stopms('Data directory not found')
c***** ensure terminater is in path
	if(lpp.eq.lp) then
	lipath=lp+1
	ipath(lipath:lipath)='/'
	else
	lipath=lp
	end if
      end if
      write(6,3)
3     format(' Give output directory : ',$)
      read(5,2) opath
      lo=len_trim(opath)
      if(lo.gt.0.and.opath(lo:lo).ne.'/') then
        lo=lo+1
        opath(lo:lo)='/'
      end if	
c***** check writeability      
      outfil='tst.tst'
      lf=7
      if(lo.ne.0) then
	if(ideb.gt.1) write(*,*)' OUTPUT to',lo,opath
	lopath=lo
        outfil=opath(1:lo)//'tst.tst'
        lf=lo+7
      end if
c***** ensure this is writeable
      if(ideb.gt.1)write(*,*)' test of file write ',lf,outfil
      open(2,file=outfil(1:lf),status='unknown')
      write(2,4,err=15) opath
4     format(a)
      close(2,status='delete')
      go to 20
15    write(6,16) opath
16    format(' Cannot write output files in directory ',a)
      call stopms('END')
20    write(6,21)
21    format(' Give first and last run numbers to treat : ',$)
      read(5,11) lin
c***** now read selection file first try local then SAS_DIR
c      read in default names,indexes,multipliers,default_values
      lin(70:79)='0 0 0 0 /'
      read(lin,*) irst,last
29    if(irst.eq.0) call stopms('nothing done')
      if(last.le.irst) last=irst
c***** look for index file (if not alread done)
      call opsdfl(35,idxfil(1:li),ier)
c      numor=34100
      do 90 kk=irst,last
      iraw=iraw+1   
      numor=kk
      write(fname,10) numor
10    format(I6.6)
      fullnam=fname
      if(lo.gt.0)fullnam=opath(1:lo)//fname
      lf=lo+6
      if(ideb.gt.1) write(*,*) ' Fullname of output: ',fullnam(1:lf)
      inquire(file=fullnam(1:lf),exist=found)
      if(found) ifound=ifound+1
c***** skip if already present - no more - overwrite
      do i=1,512
      text(i)=' '
      end do
      ix=128
      iy=128
      call sasanx(ier,numor,nomexp,ddate,nt,text,np,par,nd,idon)
      if(ideb.gt.1)write(*,*) 'ier,nt,np,nd',ier,nt,np,nd
      if(ier.ne.0.and.narg.le.1) write(*,*) ' Error reading ',numor
      if(ier.ne.0) go to 90
c****** continutng
       if(ideb.gt.1) write(*,*) 'ier=0>>>>>>>>>continuing'
c***** detector distance - calculated
      if(par(26).eq.0) par(26)=par(19)
      if(parnxs(1).ne.0) par(26)=par(19)+parnxs(1)
      if(parnxs(2).ne.0) par(26)=par(19)+parnxs(2)
      if(ideb.gt.1) write(*,*)'s-det distances p19,26,parnxs',
     1par(19),par(26),parnxs(1),parnxs(2)
      isu=0
c***** sum data array
      do i=1,nd
       isu=isu+idon(i)
      end do
      if(isu.eq.0.and.narg.lt.1) 
     1write(*,*)' WARNING - 2D data all 0 - check DNNindex.nxf file'
c***** check to see if returned raw data exceed capacity     
      if(nd.gt.ndx) call p3back(idon,iv,ix,iy)
      nd=ix*iy
      if(ideb.gt.1) write(*,*)'ier - skip...',ier
      if(ier.ne.0.and.narg.lt.1) write(*,*) 'Run not found ',numor 
      if(ier.ne.0) go to 90
c***** write out as ascii file o verwriting previous files
      if(ideb.gt.1) write(*,*)'Output file is ',lf,fullnam(1:lf)
      open(unit=1,file=fullnam(1:lf),status='unknown',err=99)
      write(1,17)
17    format(10('RRRRRRRR')) 
      iver=1
      itxt=1
      write(lin,22) numor,itxt,iver
22    format(2x,i6.6,9i8)
      write(1,11) lin
      write(lin,80) numor
80    format('From ',i6.6,'.nxs by nxs2as-1.2 and ')
      lin=lin(1:40)//version(1:40)
      write(1,11) lin
      write(1,13)
13    format(10('AAAAAAAA'))
      lin=' '
12    format(10i8)
      nc=80
      itxt=0
      write(lin,12) nc,itxt
      write(1,11) lin
      lin=instx//instx//'      '//ddate
      write(1,11) lin
11    format(a80)
c***** medpar block
      call iout(156,medpar)
      call aout(512,text)
      call fout(np,par)
      call iout(256,i0)
      write(1,18)
18    format(10('SSSSSSSS'))
c      write(*,*) ' number of data ',nd
      lin=' '
c****** for spectrum write ispec,nrest,ntot,nrun,itxt
      ispec=1
      nrest=0
      nto=1
      nrun=1
      itxt=0
      write(lin,12) ispec,nrest,nto,nrun,itxt
      write(1,11) lin
      
      call iout(nd,idon)
      close(unit=1)
      open(unit=1,file='nxs2as.log',status='unknown',
     1position='append')
      itran=itran+1
      write(1,*)'Run ',numor,'  transcribed'
      close(unit=1)
      if(narg.lt.1) write(*,*) 'Run ',numor,' transcribed '
90    continue
      if(narg.lt.1)
     1write(*,*)'Runs sought/transcribed ',iraw,itran
      call stopms('finished') 
99    write(*,*) fullnam
      end
      subroutine iout(n,iv)
      integer iv(n)
      character lin*80,iii*80
      write(1,13) 
13    format(10('IIIIIIII'))
      lin=' '
      itxt=0
      write(lin,1) n,itxt
      write(1,11) lin
      lin=' '
      nrest=mod(n,10)
      nn=n-nrest
      
      write(1,1) (iv(i),i=1,nn)
1     format(10i8)
      write(lin,1) (iv(j),j=nn+1,n)
      write(1,11) lin
11    format(a)
      return
      end

      subroutine fout(n,v)
      real v(n)
      character lin*80
      character*16 pt(105)
      data pt/
     1'        Preset 1',
     1'        Preset 2',
     1'     Time (s/10)',
     1'   Det_sum cou_2',
     1'        Mon1_sum',
     1'        Mon2_sum',
     1' Cross section W',
     1' Cross section H',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'   Det_angle deg',
     1'           BY mm',
     1'           BX mm',
     1'  changer STR mm',
     1'  det.dist DET m',
     1' Sel.speed 1/min',
     1'  Chopper1 1/min',
     1'  Chopper2 1/min',
     1'  dig.Voltmeter1',
     1'  dig.Voltmeter2',
     1' Selector A-B-C-',
     1'   Det_dist calc',
     1'     Run cycle #',
     1'    Run serial #',
     1'   Temp set pt K',
     1'    Regul temp K',
     1'   Sample temp K',
     1'  T_contr_off/on',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'  B_centr X0 pix',
     1'  B_centr Y0 pix',
     1'    Wavelength A',
     1'  Wavelength_res',
     1'  Det_pixel X mm',
     1'  Det_pixel y mm',
     1'                ',
     1'   Collimation m',
     1'        Att_type',
     1'       Att_value',
     1'                ',
     1'                ',
     1'   Sel_angle deg',
     1'          Try mm',
     1'   Sam_angle deg',
     1'   Changer pos #',
     1'   Sam_height mm',
     1'   Sc_ap_left mm',
     1'  Sc_ap_right mm',
     1'  Sc_ap_lower mm',
     1'  Sc_ap_upper mm',
     1'        GAP (mm)',
     1'    SAP W/D (mm)',
     1'    SAP H/0 (mm)',
     1'   B_stop wi/dia',
     1'    B_stop ht mm',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'     Mag field T',
     1'   Sam_pressu Pa',
     1'  React_power MW',
     1'                ',
     1'                ',
     1'                ',
     1'                ',
     1'   Pres_flag m/t',
     1' TOF curr_fram #',
     1'  Datafile vers#',
     1'                ',
     1'           ADC 1',
     1'           ADC 2',
     1'           ADC 3',
     1'           ADC 4',
     1'           DAC 1',
     1'           DAC 2',
     1'           DAC 3',
     1'           DAC 4',
     1'          Trx mm',
     1'     S_gonio deg',
     1'                ',
     1'                ',
     1'                '/
      write(1,13)
13    format(10('FFFFFFFF')) 
      lin=' '
      itxt=26
      write(lin,12) n,itxt
12    format(10i8)
      write(1,11) lin
c***** write out paras 21 lines 5 blanks
      write(1,15) pt
15    format(5a16)
      lin=' '
      do i=1,5
      write(1,16) lin
      end do
16    format(a)
      nrest=mod(n,5)
      nn=n-nrest
      write(1,1) (v(i),i=1,nn)
1     format(5e16.5)
      write(lin,1) (v(j),j=nn+1,n)
      write(1,11) lin
11    format(a)
      return
      end
      subroutine aout(n,c)
      character*1 c(n)
      character lin*80
      write(1,13) 
13    format(10('AAAAAAAA'))
      lin=' '
      itxt=0
      write(lin,12) n,itxt
12    format(10i8)
      write(1,11) lin
      lin=' '
      nrest=mod(n,80)
      nn=n-nrest
      write(1,1) (c(i),i=1,nn)
1     format(80a1)
      write(lin,1) (c(j),j=nn+1,n)
      write(1,11) lin
11    format(a)
      return
      end

      subroutine p3back(iv,ivv,ix,iy)
c***** packs as 5x5 TOF frames of 64x64 pixels
      integer iv(320,320),ivv(100,128,256)
      integer ivvv(25,128,64),iww(25,64,64)
      
c***** pack back starting at short wavelengths top left hand corner
      call p25(ivv)
c**** fill in vvv
      call p256(ivvv,ivv)
c**** reduce and transpose
      call p128(iww,ivvv)
c**** set up 25 images
      call pim(iv,iww)
      ix=320
      iy=320
      return
      end
      subroutine p25(iv)
c***** reduce from 100 to 25 time channels
      parameter (nxy=100*128*256)
      integer iv(nxy)
      n=0
      do i=1,nxy-3,4
      n=n+1
      isum=iv(i)+iv(i+1)+iv(i+2)+iv(i+3)
      iv(n)=isum
      end do
      return
      end
      subroutine p256(ivv,ivvv)
c***** reduce x to 64
      integer ivv(25,128,64),ivvv(25,128,256)
      do it=1,25
      do iy=1,128
      ixx=1
      do ix=1,64
      is=ivvv(it,iy,ixx)+ivvv(it,iy,ixx+1)+
     1ivvv(it,iy,ixx+2)+ivvv(it,iy,ixx+3)
      ivv(it,iy,ix)=is
      ixx=ixx+4
      end do      
      end do
      end do
      return
      end
      subroutine p128(ivv,ivvv)
c**** reduce y from 128 to 64 AND transpose
      integer ivv(25,64,64),ivvv(25,128,64)
      do it=1,25
      do ix=1,64
      iyy=1
      do iy=1,64
      is=ivvv(it,iyy,ix)+ivvv(it,iyy+1,ix)
      iyy=iyy+2
      ivv(it,ix,iy)=is
      end do
      end do
      end do
      return
      end
      subroutine pim(iv,iw)
c**** transfers images to 320x320 as 25 64x64 sets
      integer iv(320,320),iw(25,64,64)
c**** v is (ix,iy) in pixels, wm is (time,ix,iy)
c***** start bottom lh corner
      n=0
      ixx=4*64
      iyy=0
      do it=25,1,-1
      do iy=1,64
      do ix=1,64
      n=n+1
      iv(ixx+ix,iyy+iy)=iw(it,ix,iy)
      end do
      end do
      ixx=ixx-64
      if(ixx.lt.0) iyy=iyy+64
      if(ixx.lt.0) ixx=4*64
      end do
      return
      end
      subroutine stopms(msg)
c***** for prop on MS pause before closing window
      character*(*) msg
      character*20 os
      if(msg.eq.' ') stop
      write(6,*) msg
      call getenv('OS',os)
      if(os.eq.'Windows_NT') call sleep(5)
      stop
      end
