      subroutine sasanx(ierr,numor,nomexp,ddate,nt,text,np,
     1par,nd,idon)
c Special cache version for splots and pdqheads (xplots)
c From sasas
c     reading standard SANS ascii files, or D16 data files...
c     which may be compressed.. (*.Z files)
c     ier             integer      out  0 ok, else -1
c     numor           integer      in   1<=numor<=999999
c     inst            character*4  out  ints field in file
c     nomexp          character*10 out  nomexp in file
c     ddate           character*20 out  data and time of recording
c     nt              integer      out  number of text (a1)
c     text            character*1  out  text array
c     np              integer      out  number of real parameters
c     par             real         out  real parameters
c     nd              integer      out  number of integers in spectrum
c     idon            integer      out  spectrum
c environment variables
c  SAS_INSTRUMENT     instrument name 4 characters
c  SAS_DATA_PATH      directory for raw data
c  SAS_DEBUG    flag to print debugging information
c  SAS_DECOMPRESS zcat  "zcat" is program name of decompresion tool
c     R. E. Ghosh  March 1996   ron@ill.fr
c  SAS_CACHE     see note October 2010
c  the commons:
c  /maxarr/ ntitl,npar,nparx,ndx  512,128,20,16384 respectively
c           protect programs from excessive input. These stored 
c           defaults may be overwritten by calling programs
c  /sasasc/ nunit  file input unit is set to 31 by default
c
c  /liminp/ ininp  is set to -1 by default. If positive then only
c           that number of data fields are read-in (now there are 5) 
c  parnxs common can contain upt o 100 values selected from data file
c
c  To cope with the variety of nxs files and the placement of parameters with 
c  several possible names, this version reads a selection file which now may
c  contain duplicate entries.  The routine scannxs searches the current data
c  file and uses the appropriate information from the selection file.
c  The default value from the last read selection is initially used.  Hence
c  using a name like Default last will plant this value even if no /entry0
c  dataset name in found.
c
c  An additional feature is to allow a direcory path other than SAS_DATA_PATH
c  to be used for raw data; this helps the conversion program nxs2as
c  write into a local SAS_DATA_PATH having read an alternative raw data source
c  directory.
c
c Example: D33index.nxf
c# D33 v1.0  16-Sep-2012   Ron Ghosh
c# please modify version and date at each edition
c# scannxs - shows datasets in .nxs files 
c#
c# 1    preset 1
c# 3    time(0.1 secs)
c# 4    total detector
c# 5    total monitor
c# 23   dvm    
c# 26   s-det_m
c# 29   temperature-set
c# 30   regul-temperature
c# 31   sample-temperature
c# 51   detector_centre_x_address
c# 52   detector_centre_y_address
c# 53   wavelength
c# 54   delta_wavelength
c# 55   dx_pixel
c# 56   dy_pixel 
c# 15   two_theta(gamma)
c# 65   omega
c# 64   trs_translation
c# 200  additive to nominal detector distance
c# 201  additive to nominal detector distance
c# 300  start_time
c# 301  end_time
c# 310  run_number
c#
c# 401  sample_description
c# 400  experiment_title
c#
c# 500  data_2d..reduced to 128x128 in idon as IDAT(ix,iy)
c#      left in common/h5work/id(3276800) as original
c# Note data are read last after parameters - dx,dy may thus be modified
c
c pixel size is so critical that it is hardwird here...
c
c
c
c/entry0/D33/detector/detsum
c/entry0/D33/detector/pixel_size_x   55
c/entry0/D33/detector/pixel_size_y   56
c/entry0/D33/selector/wavelength   53
c/entry0/D33/selector/wavelength_res  54
c/entry0/sample/sdi1_actual     200
c/entry0/sample/sdi2_actual     201
c/entry0/data1/data1        533
c/entry0/duration          3     10
c/entry0/end_time          301
c/entry0/experiment_title  400
c/entry0/sample_description 401  
c/entry0/start_time            300
c/entry0/run_number   310
c/entry0/sample/temperature        31
c/entry0/sample/trs_actual         64
c/entry0/monitor2/preset    1
c/entry0/data1/data1        500

      parameter (NXYT=100*128*256)
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasnxc/version
      common/sasasc/nunit
      common/nx2asl/localin,localout
      common/sasmed/medpar(156)
      common/sasinx/instx
      common/liminp/ininp
      common/devfil/fdev
      common/devfii/lf
      common/sdbg/ideb
      common/h5ctl/ipack
      common/nxpars/parnxs(100)
      common/h5loc/locs(200)
      common/h5locl/locl(200),nloc
      common/h5lpar/locp(200)
      common/h5locv/vdef(200)
      common/h5mulv/vmul(200)
      common/h5work/vv(NXYT)
      common/h5workk/v(1024)
      common/h5lims/idmax,namax,ioneh5,ioph5
      common/scannam/dnames(500)
      common/scannml/nnam,nl(500)
      common/pathc/ipath,opath
      common/pathl/lipath,lopath
      real par(128),v,vv
      integer idon(*),idim(10)
      character*8 rtype,gname
      integer ivv(NXYT),iv(1024)
      character inst*4,instx*4,nomexp*10,ddate*20,lin*140
      character fnum*6,fdev*66,errmsg*16,gdev*60,debug*20
      character sasdir*60,sasel*100,spath*100,locs*100,selfil*100
      character fname*100,stime*20,etime*20,stit*20,vc*150,ctext*150
      character dnames*100,ipath*100,opath*100
      character*1 text(512),ttext(150),version*80
      character exptit*60
      logical found,localin,localout
      equivalence(v,vc,iv),(ctext,ttext(1)),(vv,ivv)
      data ionce/0/
      save 
      ierr=-1
      if(ionce.ne.0) go to 10
      ivers=1
c***** get instrument
      ideb=0
      debug=' '
      call getenv('SAS_DEBUG',debug)
      ideb=index(debug,' ')
      inst=' '
      call getenv('SAS_INSTRUMENT',inst)
      if(inst.eq.' ') stop 'SAS_INSTRUMENT not SET'
      if(inst(1:1).eq.'d') inst(1:1)='D'
      spath=' '
c***** allow for SAS_DATA_PATH or alternative (eg set in nxs2as)
c***** lipath,lopath are -1 for local directory
      lp=0
      if(lipath.gt.0) spath=ipath
      if(lipath.gt.0) lp=lipath
      if(lipath.eq.0) then
       call getenv('SAS_DATA_PATH',spath)
       if(spath.eq.' ') stop 'Input data directory not SET'
       lp=len_trim(spath)
      end if
      sasdir=' '
      selfil='DNNindex.nxf'
      selfil(2:2)=inst(2:2)
      selfil(3:3)=inst(3:3)
      call opsdfl(36,selfil,ier)
c***** look first for local file rhen in SAS_DIR
      if(ier.ne.0)stop 'No data file deescriptor file DNNindex.nxf'
10    continue
c
C***** read data from file and use default/multiplierc
c           
      if(ideb.ne.1) write(*,*)'SASNX input path/',lipath,ipath,':',spath
      do i=1,1000
        iv(i)=0.
	idon(i)=0
      end do
      do i=1,512
      text(i)=' '
      end do
      do i=1,156
      medpar(i)=0
      end do
      do i=1,128
      par(i)=0.
      end do
      do i=1,100
      parnxs(i)=0.
      end do
      etime=' '
      stime=' '
      ddate=' '
      stit=' '
      exptit=' '
      fname=' '
      if(lp.gt.0) fname=spath

      ctext=' '
      stit=' '
      ndind=0
      nnumor=0
      nd=0
      do i=1,128
      par(i)=0.
      end do
      np=128
      write(fname(lp+1:100),40) numor
c      write(*,*) ' numor',numor      
      lf=lp+10
40    format(i6.6,'.nxs') 
      if(ideb.ne.1) write(*,*)' Input file',lf,fname(1:lf)
      inquire(file=fname(1:lf),exist=found)
      if(ideb.ne.1)  write(*,*) ' seek fname ',found,fname     
      if(.not.found) return

      call scanxs(fname(1:lf),'   ')
      if(ideb.ne.1) write(*,*)' scannxs has found ',nnam,' names' 
c***** get current file names
c scannxs names start with //
      do i=1,nnam
      kk=nl(i)
      k=kk-1
      dnames(i)(1:k)=dnames(i)(2:kk)
      nl(i)=k
      end do   
      
c***** have index in memory from scannxs      
      
c***** remove duplicates from selection file using actual data names
      
      call oldh5(ier,id_f,fname(1:lf),'O')
      
      do 80 i=1,nloc
c***** for each data element look for matching name
c      first set default value if it is a (1-100) parameter
c      write(*,*)' seeking match locl',locl(i)      
      idim(1)=1024
      if(locp(i).eq.0) go to 80
      jj=locp(i)
c***** take default value from first occurrence if set
      if(jj.le.128.and.vdef(i).ne.0.and.par(jj).eq.0)
     1par(jj)=vdef(i)
c      write(*,*) jj,par(jj),vdef(i)
c***** set default value
      if(locp(i).le.100) xmul=vmul(i)
c***** default multiplier
c      write(*,*)' number of file names ',nnam
      if(jj.le.128.and.par(jj).ne.0.) go to 80
c***** skip rest of search

      do 70 j=1,nnam
c      write(*,*) ' matching lengths',nl(j),locl(i)

      if(nl(j).ne.locl(i)) go to 70
c***** wrong name length    
      mno=locl(i)
      if(dnames(j)(1:mno).eq.locs(i)(1:mno)) go to 75
70    continue
c      if(ideb.ne.1)write(*,*)' match',j,i,mno,dnames(j)(1:nmo),
c     1'  ',locs(i)(1:nmo)
c***** arrive here - no match
      go to 80

c***** name found - continue to read data
75    continue
      if(ideb.ne.1) write(*,*)'Treating input',i,locs(i)(1:mno)
      if(locl(i).gt.0.and.locp(i).lt.500) 
     1call readds(ier,id_f,locs(i)(1:locl(i))
     1,v,ir,idim,itype)
c***** data array would create an error (size greater than 1024)
      if(ideb.ne.1.and.ier.ne.0) 
     1write(*,*) 'Error reading ',locs(i)(1:locl(i))
      ntot=idim(1)
      if(ir.gt.1) ntot=ntot*idim(2)
      if(ir.gt.2) ntot=ntot*idim(3)
      if(ir.gt.3) ntot=ntot*idim(4)
      if(ideb.ne.1.and.itype.eq.0) 
     1write(*,*) 'par#  ntot,ir, v1 ',locp(i),ntot,ir,vc(1:ntot)
      if(ideb.ne.1.and.itype.eq.1) 
     1write(*,*) 'par#  ntot,ir, v1 ',locp(i),ntot,ir,iv(1)
      if(ideb.ne.1.and.itype.eq.2) 
     1write(*,*) 'par#  ntot,ir, v1 ',locp(i),ntot,ir,v(1)
      if(locp(i).gt.128) go to 49
c***** treat special cases 
      if(xmul.eq.0) xmul=1.
      if(itype.eq.2)par(locp(i))=v(1)*xmul
      if(itype.eq.1)par(locp(i))=iv(1)*xmul
      go to 50
49    continue
      if(locp(i).gt.299) go to 59
c***** parnxs now 1-100 specifically for nxs parameters
      jj=locp(i)-200
      if(itype.eq.2) parnxs(jj)=v(1)*xmul
      if(itype.eq.1) parnxs(jj)=iv(1)*xmul
      if(ideb.ne.1) 
     1write(*,*) 'parnxs#  ntot,ir, v1 ',jj,ntot,ir,v(1)
      go to 50
59    if(locp(i).eq.300) stime=vc(1:ntot)
      if(locp(i).eq.301) etime=vc(1:ntot)
      if(locp(i).eq.301) ddate=etime      
      if(locp(i).eq.310) nnumor=iv(1) 
c***** strip out NOMADs non-printing characters
      if(locp(i).eq.400.or.locp(i).eq.401) then
       do k=1,ntot
       if(ichar(vc(k:k)).lt.32) vc(k:k)=' '
       end do
      end if
c***** truncates to 60 clean chars
      if(locp(i).eq.400) exptit=vc(1:ntot)
c***** memo text is ch*1  fill with title /dataes etc
      if(locp(i).eq.401) stit=vc(1:ntot)
      if(locp(i).ge.500.and.ideb.ne.1) 
     1write(*,*) '2D-data locp i ',locp(i),i,locs(i)
      if(locp(i).ge.500) ndind=i
      if(locp(i).ge.500) ndold=locp(i)
c***** treat later
50    continue

80    continue
c***** intention is to leave raw data in array vv, hence this is read last
      iii=ndind
      idim(1)=NXYT
c***** can expand without recompiling
      if(idmax.gt.NXYT) idim(1)=idmax
      if(iii.gt.0) call readds(ier,id_f,locs(iii)(1:locl(iii))
     1,vv,ir,idim,itype)
      if(ideb.ne.1) 
     1write(*,*) 'nxsdat#  ntot,ir,idim ',i,ntot,ir,(idim(j),j=1,ir)
      ntot=0
      if(ier.ne.0) call oldh5(irrr,id_f,fname,'C')
      if(ier.ne.0) return
      ntot=idim(1)
      if(ir.gt.1) ntot=ntot*idim(2)
      if(ir.gt.2) ntot=ntot*idim(3)
      if(ir.gt.3) ntot=ntot*idim(4)
      npakd=0
      if(ndold.eq.512.and.ideb.gt.1) 
     1write(*,*) 'Old d22 data - do not transpose'
c***** D11 2014 transpose
      if(inst(1:3).eq.'D11'.and.ntot.eq.16384) call pack11(idon,ivv)
      if(inst(1:3).eq.'D11'.and.ntot.eq.65536) then
       call pack114(idon,ivv,par)
       ntot=16384
      end if
c***** D22 transfer from 256x128 (y,x)to 128x128(x,y)
c
c      note do 16k first then 32k otherwise double trouble!
      if(inst(1:3).eq.'D22'.and.ntot.eq.16384
     1.and.ndold.eq.500)call pack221(idon,ivv)
      if(inst(1:3).eq.'D22'.and.ntot.eq.32768
     1.and.ndold.eq.500) then
       call pack22(idon,ivv)
       ntot=16384
      end if
c***** D22 transfer directly
      if(inst(1:3).eq.'D22'.and.ntot.eq.16384
     1.and.ndold.eq.512)call pack2212(idon,ivv)
c***** 128x128
      if(inst(1:3).eq.'D22'.and.ntot.eq.32768
     1.and.ndold.eq.512) then
       call pack2222(idon,ivv)
       ntot=16384
      end if
c***** 128x128
      if(inst(1:3).eq.'D22'.and.ntot.eq.32768
     1.and.ndold.eq.522) then
       call pack222(idon,ivv)
       ntot=16384
      end if
c      128x x  256y
c***** D33 reduce and transpose
      if(inst(1:3).eq.'D33'.and.ntot.eq.32768) then 
       call pack33(idon,ivv,ntot)
       ntot=16384
      end if
c***** for other cases (eg D33-tof) data remain in common and IDON is zeros
      if(locp(iii).ge.500) nnn=ntot 
      if(locp(iii).ge.500) nd=ntot
c***** other case- returns full array in original form
c***** medpar 
      medpar(149)=3
      ddate=etime
      ctext=' '
      ctext=exptit//stit
      ctext(81:100)=stime
      ctext(101:120)=etime
c***** write 150 characters back into text array
      nt=512
      do i=1,150
      text(i)=' '
      if(ichar(ttext(i)).gt.0) text(i)=ttext(i)
      end do
      
c***** set pixel size here  dx=55 dy=56 for output arrays 128x128
      if(inst.eq.'D11') par(55)=7.5
      if(inst.eq.'D11') par(56)=7.5
      if(inst.eq.'D22') par(55)=8.0
      if(inst.eq.'D22') par(56)=8.0
      if(inst.eq.'D33') par(55)=5.0
      if(inst.eq.'D33') par(56)=5.0

      call oldh5(ier,id_f,fname,'C')
      if(ideb.ne.1) write(*,*) 'Numor seek/found',numor,nnumor
      if(nnumor.ne.numor) write(*,*) ' SASANX - run number mismatch'
      if(nnumor.ne.numor) return
      ierr=0
      return
      end

      subroutine pack22(idon,iv)
c***** for d22 packs pixels AND transposes (vertical tubes pair y values)
c      (d33 is horiz)
      common/sdbg/ideb
      integer idon(128,128),iv(32768) 
c***** first pack y then transpose
      integer iw(128,128),iww(16384)
      equivalence(iw(1,1),iww(1))
      if(ideb.ne.0) write(*,*) 'D22 reduce & transpose'
      n=1
c**** pair y values
c      write(*,*) (iv(i),i=1,10)
      do i=1,16384
      iww(i)=iv(n)+iv(n+1)
      n=n+2
      end do   
c***** transpose      
c      write(*,*) (iww(i),i=1,10)
      do iy=1,128
      do ix=1,128
      idon(ix,iy)=iw(iy,ix)
      end do
      end do
c      write(*,*) 'x',(idon(i,64),i=1,128)
c      write(*,*) 'y',(idon(64,i),i=1,128)
      
      return
      end

      
      subroutine pack33(idon,iv)
c***** for d33 packs as 5mm by 5mm pixels AND transposes
      integer idon(128,128),iv(128,256)
c***** collapsed time is first dimension ==1
      do iy=1,128
      ixx=0
      do ix=1,128
      sum=0.
      ixx=ixx+1
      sum=iv(iy,ixx)
      ixx=ixx+1
      idon(ix,iy)=iv(iy,ixx)+sum
      end do
      end do
      ntot=ntot/2
c      write(*,*) ' pack2 par',par
c      write(*,*)' ntot ',ntot
      return
      end
      subroutine pack114(ivv,ivvv)
c**** reduce from 256x256 to 128x128 and transpose
c
      integer ivv(128,128),ivvv(256,256)
c      write(*,*)' D11 256x256 data reduced to 128x128'
      iyy=0
      ixx=0
      do iy=1,255,2
      iyy=iyy+1
      ixx=0
      do ix=1,255,2
      ixx=ixx+1
      ivv(iyy,ixx)=ivvv(ix,iy)+ivvv(ix+1,iy)+
     1ivvv(ix,iy+1)+ivvv(ix+1,iy+1)
      end do
      end do
      ix=128
      iy=128
      return
      end

      subroutine pack11(ivv,ivvv)
c***** transposes 12x128
c
      integer ivv(128,128),ivvv(128,128)
c      write(*,*)' D11 transposed'
      iyy=0
      ixx=0
      do iy=1,128
      do ix=1,128
      ivv(iy,ix)=ivvv(ix,iy)
      end do
      end do
      ix=128
      iy=128
      return
      end

      subroutine pack221(ivv,ivvv)
c***** transposes 12x128
c
      integer ivv(128,128),ivvv(128,128)
c      write(*,*)' D22  128x128  transposed'
      iyy=0
      ixx=0
      do iy=1,128
      do ix=1,128
      ivv(iy,ix)=ivvv(ix,iy)
      end do
      end do
      ix=128
      iy=128
      return
      end

      subroutine pack222(ivv,ivvv)
c**** reduce from 128x256 to 128x128  X..Y
c
      integer ivv(128,128),ivvv(128,256)
c      write(*,*)' D22 128x256 data reduced to 128x128; transposed'
      iyy=0
      ixx=0
      do iy=1,255,2
      iyy=iyy+1
      ixx=0
      do ix=1,128
      ixx=ixx+1
      ivv(ixx,iyy)=ivvv(iy,ix)+ivvv(iy+1,ix)
      end do
      end do
      ix=128
      iy=128
      return
      end
      
      subroutine pack2212(ivv,ivvv)
c***** transposes 12x128
c
      integer ivv(128,128),ivvv(128,128)
c      write(*,*)' D22  128x128  untransposed'
      iyy=0
      ixx=0
      do iy=1,128
      do ix=1,128
      ivv(ix,iy)=ivvv(ix,iy)
      end do
      end do
      ix=128
      iy=128
      return
      end

      subroutine pack2222(ivv,ivvv)
c**** reduce from 128x256 to 128x128  X..Y
c
      integer ivv(128,128),ivvv(128,256)
c      write(*,*)' D22 128x256 data reduced to 128x128; untransposed'
      iyy=0
      ixx=0
      do iy=1,255,2
      iyy=iyy+1
      ixx=0
      do ix=1,128
      ivv(ix,iyy)=ivvv(ix,iy)+ivvv(ix,iy+1)
      end do
      end do
      ix=128
      iy=128
      return
      end
      
      subroutine opsdfl(nunit,sasfil,ier)
c***** reads index file DNNindex.nxf for sought values and defaults
      common/sdbg/ideb
      common/h5loc/locs(200)
      common/h5locl/locl(200),nloc
      common/h5lpar/locp(200)
      common/h5locv/vdef(200)
      common/h5mulv/vmul(200)
      common/indexc/findnam(100)
      common/indexi/ifind(100),loc
      common/indone/ionce
      common/sasinx/instx
      common/sasnxc/version
      character*(*) sasfil
      character*100 sfil,sasdir*100,findnam*100,locs*100,lin*130
      character debug*20
      character version*80
      logical found
      ier=0
      if(ionce.ne.0) return
      sfil=sasfil
      ls=len_trim(sfil)
      ier=-1
      ideb=0
      debug=' '
      sasdir=' '
      call getenv('SAS_DEBUG',debug)
      ideb=index(debug,' ')
      call getenv('SAS_DIR',sasdir)
      lsd=len_trim(sasdir)
      if(lsd.gt.0) lsd=lsd+1
      if(lsd.gt.0)sasdir(lsd:lsd)='/'
c***** name of selection-help file depends on instrument
      inquire(file=sfil,exist=found)
      if(.not.found.and.lsd.gt.0) sfil=sasdir(1:lsd)//sfil
      version=sfil
      ls=len_trim(sfil)
      inquire(file=sfil,exist=found)
      if(ideb.ne.1) write(*,*) 'SASNX  index ',sfil(1:ls),found
      if(.not.found) return
      open(unit=nunit,file=sfil,status='old',err=20)
c***** initialise
      do i=1,100
      findnam(i)=' '
      ifind(i)=0
      vdef(i)=0
      vmul(i)=0.
      end do
      nloc=1
      do i=1,500
c***** read index datasetname multiplier default value      
      read(nunit,1,end=14)lin
c      write(*,*)lin
1     format(a)
      if(lin(1:1).eq.'#') go to 10
c***** skip comments
      jj=index(lin,' ')-1
      lloc=jj
      locl(nloc)=jj
      locs(nloc)=lin(1:lloc)
      ifind(nloc)=ind
      ifind(loc)=lloc
      lin(1:jj)='            '
      lin(120:)=' 0 0 0 0/'
      read(lin(jj:),*,err=99) locp(nloc),vmul(nloc),vdef(nloc)
      if(locp(nloc).le.0) go to 8
      if(ideb.gt.1.and.locp(nloc).gt.0)
     1write(*,*) 'saving...location',locs(nloc)(1:lloc), locp(nloc)
      nloc=nloc+1
8     continue
      go to 10
99    if(ideb.gt.1) write(*,*) 'Error reading line ',i,lin
10    continue
      
      end do  
      nloc=nloc-1
      close(nunit)      
14    continue    
      if(ideb.ne.1) write(*,*)'SASNX number of indexed components',nloc
      ier=0
      ionce=1
      return
20    if(ideb.ne.1) write(*,*) 'SASNX indexfile DNNindex.nxf missing' 
      return
      end
c***** beware of mixing with sasas - already has these common blocks...
      block data sasasb
      common/maxarr/ntitl,npar,nparx,ndx
      common/sasasc/nunit
      common/liminp/ininp
      common/indone/ionce
      common/pathl/lipath,lopath
      COMMON/NEWMASK/GRDD
      save
      character*8 GRDD
      data GRDD/'UNMASKED'/
c***** SASAS is always present when masking is employed
      data ntitl/512/,npar/128/,nparx/20/,ndx/102400/
      data nunit/31/
      data ininp/-1/
      data ionce/0/
      data lipath/0/
      data lopath/0/
      end
